import autora.synthetic
from autora.synthetic.inventory import SyntheticExperimentCollection
from autora.variable import VariableCollection


def test_model_registration_retrieval():
    # We can register a model and retrieve it
    autora.synthetic.register("empty", lambda: SyntheticExperimentCollection())
    empty = autora.synthetic.retrieve("empty")
    assert empty.name is None

    # We can register another model and retrieve it as well
    autora.synthetic.register(
        "only_metadata",
        lambda: SyntheticExperimentCollection(metadata=VariableCollection()),
    )
    only_metadata = autora.synthetic.retrieve("only_metadata")
    assert only_metadata.metadata is not None

    # We can still retrieve the first model, and it is equal to the first version
    empty_copy = autora.synthetic.retrieve("empty")
    assert empty_copy == empty



from typing import Iterable

import numpy as np
from sklearn.metrics import mean_absolute_error as mae
from sklearn.metrics import mean_squared_error as mse


def assumption_sampler(
    X, y, model, n, loss=True, theorist=None, confirmation_bias=False
):
    """
    Assumption Sampler challenges assumptions made by the Theorist.
    It identifies points whose error are most dependent on the assumption made.
    Assumptions take the form of hard-coding, which may be hyperparameters or arbitrarily chosen
    sub-algorithms e.g. loss function
    Because it samples with respect to a Theorist, this sampler cannot be used on the first cycle

    Args:
        X: pool of IV conditions to sample from
        y: experimental results from most recent iteration
        model: Scikit-learn model, must have `predict` method.
        n: number of samples to select
        loss: assumption to test: identify points that are most affected by choice of loss function
        theorist: the Theorist, which employs the theory it has been hard-coded to demonstrate
        confirmation_bias: whether to find evidence to support or oppose the theory

    Returns: Sampled pool

    """

    if isinstance(X, Iterable):
        X = np.array(list(X))
    current = None
    if theorist:
        pass  # add code to extract loss function from theorist object
    idx = range(len(X))

    if y is not None:
        if loss:
            if current is None:
                current = mse
                print(
                    Warning(
                        "Knowledge of Theorist Loss Function needed. MSE has been assumed."
                    )
                )
            y_pred = model.predict(X)
            current_loss = current(
                y_true=y.reshape(1, -1),
                y_pred=y_pred.reshape(1, -1),
                multioutput="raw_values",
            )
            print(current_loss)
            alternative = mae
            alternative_loss = alternative(
                y_true=y.reshape(1, -1),
                y_pred=y_pred.reshape(1, -1),
                multioutput="raw_values",
            )
            loss_delta = alternative_loss - current_loss
            idx = np.flip(loss_delta.argsort()[:n])
    else:
        raise TypeError(
            "Experiment results are required to run the assumption experimentalist"
        )

    return X[idx]



from typing import Iterable, Literal

import numpy as np
from sklearn.metrics import DistanceMetric

AllowedMetrics = Literal[
    "euclidean",
    "manhattan",
    "chebyshev",
    "minkowski",
    "wminkowski",
    "seuclidean",
    "mahalanobis",
    "haversine",
    "hamming",
    "canberra",
    "braycurtis",
    "matching",
    "jaccard",
    "dice",
    "kulsinski",
    "rogerstanimoto",
    "russellrao",
    "sokalmichener",
    "sokalsneath",
    "yule",
]


def summed_dissimilarity_sampler(
    X: np.ndarray, X_ref: np.ndarray, n: int = 1, metric: AllowedMetrics = "euclidean"
) -> np.ndarray:
    """
    This dissimilarity samples re-arranges the pool of IV conditions according to their
    dissimilarity with respect to a reference pool X_ref. The default dissimilarity is calculated
    as the average of the pairwise distances between the conditions in X and X_ref.

    Args:
        X: pool of IV conditions to evaluate dissimilarity
        X_ref: reference pool of IV conditions
        n: number of samples to select
        metric (str): dissimilarity measure. Options: 'euclidean', 'manhattan', 'chebyshev',
            'minkowski', 'wminkowski', 'seuclidean', 'mahalanobis', 'haversine',
            'hamming', 'canberra', 'braycurtis', 'matching', 'jaccard', 'dice',
            'kulsinski', 'rogerstanimoto', 'russellrao', 'sokalmichener',
            'sokalsneath', 'yule'. See [sklearn.metrics.DistanceMetric][] for more details.

    Returns:
        Sampled pool
    """

    if isinstance(X, Iterable):
        X = np.array(list(X))

    if isinstance(X_ref, Iterable):
        X_ref = np.array(list(X_ref))

    if X.ndim == 1:
        X = X.reshape(-1, 1)

    if X_ref.ndim == 1:
        X_ref = X_ref.reshape(-1, 1)

    if X.shape[1] != X_ref.shape[1]:
        raise ValueError(
            f"X and X_ref must have the same number of columns.\n"
            f"X has {X.shape[1]} columns, while X_ref has {X_ref.shape[1]} columns."
        )

    if X.shape[0] < n:
        raise ValueError(
            f"X must have at least {n} rows matching the number of requested samples."
        )

    dist = DistanceMetric.get_metric(metric)

    # create a list to store the summed distances for each row in matrix1
    summed_distances = []

    # loop over each row in first matrix
    for row in X:
        # calculate the distances between the current row in matrix1 and all other rows in matrix2
        summed_distance = 0

        for X_ref_row in X_ref:

            distance = dist.pairwise([row, X_ref_row])[0, 1]
            summed_distance += distance

        # store the summed distance for the current row
        summed_distances.append(summed_distance)

    # sort the rows in matrix1 by their summed distances
    sorted_X = X[np.argsort(summed_distances)[::-1]]

    return sorted_X[:n]



"""
Provides an interface for loading and saving synthetic experiments.

Examples:
    The registry is accessed using the `retrieve` function, optionally setting parameters:
    >>> from autora.synthetic import retrieve, describe
    >>> import numpy as np
    >>> s = retrieve("weber_fechner",rng=np.random.default_rng(seed=180))  # the Weber-Fechner Law

    Use the describe function to give information about the synthetic experiment:
    >>> describe(s) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    Weber-Fechner Law...

    The synthetic experiement `s` has properties like the name of the experiment:
    >>> s.name
    'Weber-Fechner Law'

    ... a valid metadata description:
    >>> s.metadata  # doctest: +ELLIPSIS
    VariableCollection(...)

    ... a function to generate the full domain of the data (if possible)
    >>> x = s.domain()
    >>> x   # doctest: +ELLIPSIS
    array([[0...]])

    ... the experiment_runner runner which can be called to generate experimental results:
    >>> import numpy as np
    >>> y = s.experiment_runner(x)  # doctest: +ELLIPSIS
    >>> y
    array([[ 0.00433955],
           [ 1.79114625],
           [ 2.39473454],
           ...,
           [ 0.00397802],
           [ 0.01922405],
           [-0.00612883]])

    ... a function to plot the ground truth:
    >>> s.plotter()

    ... against a fitted model if it exists:
    >>> from sklearn.linear_model import LinearRegression
    >>> model = LinearRegression().fit(x, y)
    >>> s.plotter(model)

    These can be used to run a full experimental cycle
    >>> from autora.experimentalist.pipeline import make_pipeline
    >>> from autora.experimentalist.pooler.grid import grid_pool
    >>> from autora.experimentalist.sampler.random_ import random_sampler
    >>> from functools import partial
    >>> import random
    >>> metadata = s.metadata
    >>> pool = partial(grid_pool, ivs=metadata.independent_variables)
    >>> random.seed(181) # set the seed for the random sampler
    >>> sampler = partial(random_sampler, n=20)
    >>> experimentalist_pipeline = make_pipeline([pool, sampler])

    >>> from autora.cycle import Cycle
    >>> theorist = LinearRegression()

    >>> cycle = Cycle(metadata=metadata, experimentalist=experimentalist_pipeline,
    ...               experiment_runner=s.experiment_runner, theorist=theorist)

    >>> c = cycle.run(10)
    >>> c.data.theories[-1].coef_   # doctest: +ELLIPSIS
    array([-0.53610647,  0.58457307])
"""

from autora.synthetic import data
from autora.synthetic.inventory import (
    Inventory,
    SyntheticExperimentCollection,
    describe,
    register,
    retrieve,
)



import itertools
from typing import Iterable, List

import numpy as np
from numpy.typing import ArrayLike


def model_disagreement_sampler(X: ArrayLike, models: List, num_samples: int = 1):
    """
    A sampler that returns selected samples for independent variables
    for which the models disagree the most in terms of their predictions.

    Args:
        X: pool of IV conditions to evaluate in terms of model disagreement
        models: List of Scikit-learn (regression or classification) models to compare
        num_samples: number of samples to select

    Returns: Sampled pool
    """

    if isinstance(X, Iterable):
        X = np.array(list(X))

    X_predict = np.array(X)
    if len(X_predict.shape) == 1:
        X_predict = X_predict.reshape(-1, 1)

    model_disagreement = list()

    # collect diagreements for each model pair
    for model_a, model_b in itertools.combinations(models, 2):

        # determine the prediction method
        if hasattr(model_a, "predict_proba") and hasattr(model_b, "predict_proba"):
            model_a_predict = model_a.predict_proba
            model_b_predict = model_b.predict_proba
        elif hasattr(model_a, "predict") and hasattr(model_b, "predict"):
            model_a_predict = model_a.predict
            model_b_predict = model_b.predict
        else:
            raise AttributeError(
                "Models must both have `predict_proba` or `predict` method."
            )

        # get predictions from both models
        y_a = model_a_predict(X_predict)
        y_b = model_b_predict(X_predict)

        assert y_a.shape == y_b.shape, "Models must have same output shape."

        # determine the disagreement between the two models in terms of mean-squared error
        if len(y_a.shape) == 1:
            disagreement = (y_a - y_b) ** 2
        else:
            disagreement = np.mean((y_a - y_b) ** 2, axis=1)

        model_disagreement.append(disagreement)

    assert len(model_disagreement) >= 1, "No disagreements to compare."

    # sum up all model disagreements
    summed_disagreement = np.sum(model_disagreement, axis=0)

    # sort the summed disagreements and select the top n
    idx = (-summed_disagreement).argsort()[:num_samples]

    return X[idx]



import random

import numpy as np
import pytest
from sklearn.linear_model import LinearRegression
from sklearn.metrics import r2_score

from autora.cycle import (
    Cycle,
    cycle_default_score,
    cycle_specified_score,
    plot_cycle_score,
    plot_results_panel_2d,
    plot_results_panel_3d,
)
from autora.cycle.plot_utils import _check_replace_default_kw
from autora.experimentalist.pipeline import Pipeline
from autora.experimentalist.pooler.grid import grid_pool
from autora.experimentalist.sampler.random_ import random_sampler
from autora.variable import Variable, VariableCollection


@pytest.fixture
def ground_truth_1x():
    def ground_truth(xs):
        return xs + 1.0

    return ground_truth


@pytest.fixture
def cycle_lr(ground_truth_1x):
    random.seed(1)

    # Variable Metadata
    study_metadata = VariableCollection(
        independent_variables=[
            Variable(name="x1", allowed_values=np.linspace(0, 1, 100))
        ],
        dependent_variables=[Variable(name="y", value_range=(-20, 20))],
    )

    # Theorist
    lm = LinearRegression()

    # Experimentalist
    example_experimentalist = Pipeline(
        [
            ("pool", grid_pool),
            ("sampler", random_sampler),
            ("transform", lambda x: [s[0] for s in x]),
        ],
        params={
            "pool": {"ivs": study_metadata.independent_variables},
            "sampler": {"n": 5},
        },
    )

    # Experiment Runner
    def get_example_synthetic_experiment_runner():
        rng = np.random.default_rng(seed=180)

        def runner(xs):
            return ground_truth_1x(xs) + rng.normal(0, 0.1, xs.shape)

        return runner

    example_synthetic_experiment_runner = get_example_synthetic_experiment_runner()

    # Initialize Cycle
    cycle = Cycle(
        metadata=study_metadata,
        theorist=lm,
        experimentalist=example_experimentalist,
        experiment_runner=example_synthetic_experiment_runner,
    )

    # Run 10 iterations
    cycle.run(10)

    return cycle


@pytest.fixture
def ground_truth_2x():
    def ground_truth(X):
        return X[:, 0] + (0.5 * X[:, 1]) + 1.0

    return ground_truth


@pytest.fixture
def cycle_multi_lr(ground_truth_2x):
    random.seed(1)

    # def ground_truth(X):
    #     return X[:, 0] + (0.5 * X[:, 1]) + 1.0

    # Variable Metadata
    study_metadata = VariableCollection(
        independent_variables=[
            Variable(name="x1", allowed_values=np.linspace(0, 1, 10)),
            Variable(name="x2", allowed_values=np.linspace(0, 1, 10)),
        ],
        dependent_variables=[Variable(name="y", value_range=(-20, 20))],
    )

    # Theorist
    lm = LinearRegression()

    # Experimentalist
    example_experimentalist = Pipeline(
        [
            ("pool", grid_pool),
            ("sampler", random_sampler),
            ("transform", lambda x: np.array(x)),
        ],
        params={
            "pool": {"ivs": study_metadata.independent_variables},
            "sampler": {"n": 10},
        },
    )

    # Experiment Runner
    def get_example_synthetic_experiment_runner():
        rng = np.random.default_rng(seed=180)

        def runner(xs):
            return ground_truth_2x(xs) + rng.normal(0, 0.25, xs.shape[0])

        return runner

    example_synthetic_experiment_runner = get_example_synthetic_experiment_runner()

    # Initialize Cycle
    cycle = Cycle(
        metadata=study_metadata,
        theorist=lm,
        experimentalist=example_experimentalist,
        experiment_runner=example_synthetic_experiment_runner,
    )

    # Run 6 iterations
    cycle.run(6)

    return cycle


def test_check_replace_default_kw():
    default = {
        "subplot_kw": {"sharex": True, "sharey": True},
        "gridspec_kw": {"bottom": 0.16},
    }
    user = {
        "new_kw": True,
        "subplot_kw": {"sharey": False},
        "gridspec_kw": {"bottom": 0.2, "top": 0.9},
    }
    d_result = _check_replace_default_kw(default, user)

    assert d_result == {
        "subplot_kw": {"sharex": True, "sharey": False},
        "gridspec_kw": {"bottom": 0.2, "top": 0.9},
        "new_kw": True,
    }


def test_2d_plot(cycle_lr):
    """
    Tests plotting functionality of plot_results_panel_2d.
    """
    steps = 51
    fig = plot_results_panel_2d(
        cycle_lr, steps=steps, wrap=3, subplot_kw={"sharex": True, "sharey": True}
    )

    # Should have 12 axes, 10 with data and the last 2 turned off
    assert len(fig.axes) == 12
    assert sum([s.axison for s in fig.axes]) == 10

    # Check number of data points on each figure
    # Blue dots should start at 0 and augment by 5.
    # Orange should always be 5-this is the condition sampling rate set by the Experimentalist.
    l_counts = []
    for axes in fig.axes[:-2]:
        blue_dots = (
            len(axes.collections[0].get_offsets().mask)
            - axes.collections[0].get_offsets().mask.any(axis=1).sum()
        )
        orange_dots = (
            len(axes.collections[1].get_offsets().mask)
            - axes.collections[1].get_offsets().mask.any(axis=1).sum()
        )
        l_counts.append([blue_dots, orange_dots])
    assert np.array_equal(
        np.array(l_counts),
        np.array(
            [
                [0, 5],
                [5, 5],
                [10, 5],
                [15, 5],
                [20, 5],
                [25, 5],
                [30, 5],
                [35, 5],
                [40, 5],
                [45, 5],
            ]
        ),
    )

    # Test theory line is being plotted
    for axes in fig.axes[:-2]:
        assert len(axes.lines[0].get_xdata()) == steps
        assert len(axes.lines[0].get_ydata()) == steps


def test_3d_plot(cycle_multi_lr):
    """
    Tests plotting functionality of plot_results_panel_3d.
    """
    steps = 20
    fig = plot_results_panel_3d(
        cycle_multi_lr,
        steps=steps,
        view=(20, 60),
        wrap=3,
        subplot_kw=dict(figsize=(11, 8)),
    )

    # Should have 6 axes
    assert len(fig.axes) == 6
    assert sum([s._axis3don for s in fig.axes]) == 6

    # Check number of data points on each figure
    # Blue dots should start at 0 and augment by 5.
    # Orange should always be 5-this is the condition sampling rate set by the Experimentalist.
    l_counts = []
    for axes in fig.axes[:-1]:
        blue_dots = (
            len(axes.collections[0].get_offsets().mask)
            - axes.collections[0].get_offsets().mask.any(axis=1).sum()
        )
        orange_dots = (
            len(axes.collections[1].get_offsets().mask)
            - axes.collections[1].get_offsets().mask.any(axis=1).sum()
        )
        l_counts.append([blue_dots, orange_dots])
    assert np.array_equal(
        np.array(l_counts),
        np.array([[0, 10], [10, 10], [20, 10], [30, 10], [40, 10]]),
    )


def test_score_functions(cycle_lr, ground_truth_1x):
    """
    Tests the scoring functions cycle_default_score and cycle_specified_score.
    """
    X_test = cycle_lr.data.metadata.independent_variables[0].allowed_values.reshape(
        -1, 1
    )
    y_test = ground_truth_1x(X_test)

    scores_default = cycle_default_score(cycle_lr, X_test, y_test)
    scores_specified = cycle_specified_score(r2_score, cycle_lr, X_test, y_test)

    # Check scores are the expected values
    score_values = [
        0.98950589,
        0.99352993,
        0.9858365,
        0.99909308,
        0.99811927,
        0.98663153,
        0.98748396,
        0.9848339,
        0.99359794,
        0.99691326,
    ]
    assert np.array_equal(np.around(score_values, 8), np.around(scores_default, 8))

    # Results should be equal between the two functions.
    # The default scorer of the LinearRegression estimator is r2_score.
    assert np.array_equal(scores_default, scores_specified)


def test_cycle_score_plot(cycle_lr, ground_truth_1x):
    """
    Tests plotting functionality of test_cycle_score_plot with a 2D linear regression.
    """
    X_test = cycle_lr.data.metadata.independent_variables[0].allowed_values.reshape(
        -1, 1
    )
    y_test = ground_truth_1x(X_test)
    fig = plot_cycle_score(cycle_lr, X_test, y_test)

    # Should have 1 axis
    assert len(fig.axes) == 1

    # Test line is plotted correctly
    axis = fig.axes[0]
    assert len(axis.lines[0].get_xdata()) == 10
    y_values = np.array(
        [
            0.98950589,
            0.99352993,
            0.9858365,
            0.99909308,
            0.99811927,
            0.98663153,
            0.98748396,
            0.9848339,
            0.99359794,
            0.99691326,
        ]
    )
    y_plotted = axis.lines[0].get_ydata()
    assert np.array_equal(np.around(y_plotted, 8), np.around(y_values, 8))


def test_cycle_score_plot_multi_lr(cycle_multi_lr, ground_truth_2x):
    """
    Tests plotting functionality of test_cycle_score_plot with multiple linear regression cycle.
    """
    cycle_multi_lr.run(6)  # Run additional 6 times, total of 12 cycles
    X_test = np.array(
        list(grid_pool(cycle_multi_lr.data.metadata.independent_variables))
    )
    y_test = ground_truth_2x(X_test)
    fig = plot_cycle_score(cycle_multi_lr, X_test, y_test)

    # Test line is plotted correctly
    axis = fig.axes[0]
    assert len(axis.lines[0].get_xdata()) == 12
    y_values = np.array(
        [
            0.89368929,
            0.91897824,
            0.96375643,
            0.94514076,
            0.97807231,
            0.98778323,
            0.9931792,
            0.98768072,
            0.98952749,
            0.98867354,
            0.9872955,
            0.98999052,
        ]
    )
    y_plotted = axis.lines[0].get_ydata()
    assert np.array_equal(np.around(y_plotted, 8), np.around(y_values, 8))


def test_2d_plot_indexing(cycle_lr):
    """
    Test indexing of 2d plotter.
    """
    steps = 51
    fig = plot_results_panel_2d(
        cycle_lr,
        steps=steps,
        wrap=2,
        query=[0, 3, 7],
        subplot_kw={"sharex": True, "sharey": True},
    )

    # Should have 4 axes, 3 with data and the last turned off
    assert len(fig.axes) == 4
    assert sum([s.axison for s in fig.axes]) == 3


def test_2d_plot_negative_indexing(cycle_lr):
    """
    Test indexing of 2d plotter.
    """
    steps = 51
    fig = plot_results_panel_2d(
        cycle_lr,
        steps=steps,
        wrap=2,
        query=[-2, -1],
        subplot_kw={"sharex": True, "sharey": True},
    )

    # Should have 2 axes
    assert len(fig.axes) == 2
    assert sum([s.axison for s in fig.axes]) == 2

    # Should be plotting cycles 8 and 9
    assert fig.axes[0].get_children()[3].get_text() == "Cycle 8"
    assert fig.axes[1].get_children()[3].get_text() == "Cycle 9"


def test_2d_plot_slicing(cycle_lr):
    """
    Test slicing of 2d plotter using built-in slice() function.
    """
    steps = 51

    # Using Slice function
    # Cycles 0, 2, 4, 6, 8
    fig = plot_results_panel_2d(
        cycle_lr,
        steps=steps,
        wrap=3,
        query=slice(0, 9, 2),
        subplot_kw={"sharex": True, "sharey": True},
    )
    # Should have 6 axes, 5 with data and the last turned off
    assert len(fig.axes) == 6
    assert sum([s.axison for s in fig.axes]) == 5

    # Last 4 plots
    fig2 = plot_results_panel_2d(
        cycle_lr,
        steps=steps,
        wrap=3,
        query=slice(-4, None, None),
        subplot_kw={"sharex": True, "sharey": True},
    )
    # Should have 6 axes, 4 with data
    assert len(fig2.axes) == 6
    assert sum([s.axison for s in fig2.axes]) == 4


def test_2d_plot_slicing_np(cycle_lr):
    """
    Test slicing of 2d plotter using np.s_ Index Expression
    """
    steps = 51

    # Cycles 0, 2, 4, 6, 8
    fig1 = plot_results_panel_2d(
        cycle_lr,
        steps=steps,
        wrap=3,
        query=np.s_[0:9:2],
        subplot_kw={"sharex": True, "sharey": True},
    )
    # Should have 6 axes, 5 with data and the last turned off
    assert len(fig1.axes) == 6
    assert sum([s.axison for s in fig1.axes]) == 5

    fig2 = plot_results_panel_2d(
        cycle_lr,
        steps=steps,
        wrap=3,
        query=np.s_[-4:],
        subplot_kw={"sharex": True, "sharey": True},
    )
    # Should have 6 axes, 4 with data
    assert len(fig2.axes) == 6
    assert sum([s.axison for s in fig2.axes]) == 4


def test_2d_plot_plot_single(cycle_lr):
    """
    Test query of 2d plotter for a single cycle.
    """
    steps = 51

    # Using index
    fig1 = plot_results_panel_2d(
        cycle_lr,
        steps=steps,
        query=[9],
        subplot_kw={"sharex": True, "sharey": True},
    )
    assert len(fig1.axes) == 1
    assert sum([s.axison for s in fig1.axes]) == 1

    # Using slice()
    fig2 = plot_results_panel_2d(
        cycle_lr,
        steps=steps,
        query=slice(-1, None, None),
        subplot_kw={"sharex": True, "sharey": True},
    )
    assert len(fig2.axes) == 1
    assert sum([s.axison for s in fig2.axes]) == 1

    # Using np.s_ Index expression
    fig3 = plot_results_panel_2d(
        cycle_lr,
        steps=steps,
        query=np.s_[-1:],
        subplot_kw={"sharex": True, "sharey": True},
    )
    assert len(fig3.axes) == 1
    assert sum([s.axison for s in fig3.axes]) == 1



from typing import Iterable

import numpy as np
from scipy.stats import entropy


def uncertainty_sampler(X, model, n, measure="least_confident"):
    """

    Args:
        X: pool of IV conditions to evaluate uncertainty
        model: Scikit-learn model, must have `predict_proba` method.
        n: number of samples to select
        measure: method to evaluate uncertainty. Options:

            - `'least_confident'`: $x* = \\operatorname{argmax} \\left( 1-P(\\hat{y}|x) \\right)$,
              where $\\hat{y} = \\operatorname{argmax} P(y_i|x)$
            - `'margin'`:
              $x* = \\operatorname{argmax} \\left( P(\\hat{y}_1|x) - P(\\hat{y}_2|x) \\right)$,
              where $\\hat{y}_1$ and $\\hat{y}_2$ are the first and second most probable
              class labels under the model, respectively.
            - `'entropy'`:
              $x* = \\operatorname{argmax} \\left( - \\sum P(y_i|x)
              \\operatorname{log} P(y_i|x) \\right)$

    Returns: Sampled pool

    """

    if isinstance(X, Iterable):
        X = np.array(list(X))

    a_prob = model.predict_proba(X)

    if measure == "least_confident":
        # Calculate uncertainty of max probability class
        a_uncertainty = 1 - a_prob.max(axis=1)
        # Get index of largest uncertainties
        idx = np.flip(a_uncertainty.argsort()[-n:])

    elif measure == "margin":
        # Sort values by row descending
        a_part = np.partition(-a_prob, 1, axis=1)
        # Calculate difference between 2 largest probabilities
        a_margin = -a_part[:, 0] + a_part[:, 1]
        # Determine index of smallest margins
        idx = a_margin.argsort()[:n]

    elif measure == "entropy":
        # Calculate entropy
        a_entropy = entropy(a_prob.T)
        # Get index of largest entropies
        idx = np.flip(a_entropy.argsort()[-n:])

    else:
        raise ValueError(
            f"Unsupported uncertainty measure: '{measure}'\n"
            f"Only 'least_confident', 'margin', or 'entropy' is supported."
        )

    return X[idx]



import inspect
from itertools import product
from typing import Callable, List, Optional, Sequence, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import rcParams
from matplotlib.patches import Patch
from matplotlib.ticker import MaxNLocator
from numpy.typing import ArrayLike

from .simple import SimpleCycle as Cycle

# Change default plot styles
rcParams["axes.spines.top"] = False
rcParams["axes.spines.right"] = False
rcParams["legend.frameon"] = False


def _get_variable_index(
    cycle: Cycle,
) -> Tuple[List[Tuple[int, str, str]], List[Tuple[int, str, str]]]:
    """
    Extracts information about independent and dependent variables from the cycle object.
    Returns a list of tuples of (index, name, units). The index is in reference to the column number
    in the observed value arrays.
    Args:
        cycle: AER Cycle object that has been run

    Returns: Tuple of 2 lists of tuples

    """
    l_iv = [
        (i, s.name, s.units)
        for i, s in enumerate(cycle.data.metadata.independent_variables)
    ]
    n_iv = len(l_iv)
    l_dv = [
        (i + n_iv, s.name, s.units)
        for i, s in enumerate(cycle.data.metadata.dependent_variables)
    ]
    return l_iv, l_dv


def _observed_to_df(cycle: Cycle) -> pd.DataFrame:
    """
    Concatenates observation data of cycles into a single dataframe with a field "cycle" with the
    cycle index.
    Args:
        cycle: AER Cycle object that has been run

    Returns: Dataframe

    """
    l_observations = cycle.data.observations
    l_agg = []

    for i, data in enumerate(l_observations):
        l_agg.append(pd.DataFrame(data).assign(cycle=i))

    df_return = pd.concat(l_agg)

    return df_return


def _min_max_observations(cycle: Cycle) -> List[Tuple[float, float]]:
    """
    Returns minimum and maximum of observed values for each independent variable.
    Args:
        cycle: AER Cycle object that has been run

    Returns: List of tuples

    """
    l_return = []
    iv_index = range(len(cycle.data.metadata.independent_variables))
    l_observations = cycle.data.observations
    # Get min and max of observation data
    # Min and max by cycle - All IVs
    l_mins = [np.min(s, axis=0) for s in l_observations]  # Arrays by columns
    l_maxs = [np.max(s, axis=0) for s in l_observations]
    # Min and max for all cycles by IVs
    for idx in iv_index:
        glob_min = np.min([s[idx] for s in l_mins])
        glob_max = np.max([s[idx] for s in l_maxs])
        l_return.append((glob_min, glob_max))

    return l_return


def _generate_condition_space(cycle: Cycle, steps: int = 50) -> ArrayLike:
    """
    Generates condition space based on the minimum and maximum of all observed data in AER Cycle.
    Args:
        cycle: AER Cycle object that has been run
        steps: Number of steps to define the condition space

    Returns: np.array

    """
    l_min_max = _min_max_observations(cycle)
    l_space = []

    for min_max in l_min_max:
        l_space.append(np.linspace(min_max[0], min_max[1], steps))

    if len(l_space) > 1:
        return np.array(list(product(*l_space)))
    else:
        return l_space[0].reshape(-1, 1)


def _generate_mesh_grid(cycle: Cycle, steps: int = 50) -> np.ndarray:
    """
    Generates a mesh grid based on the minimum and maximum of all observed data in AER Cycle.
    Args:
        cycle: AER Cycle object that has been run
        steps: Number of steps to define the condition space

    Returns: np.ndarray

    """
    l_min_max = _min_max_observations(cycle)
    l_space = []

    for min_max in l_min_max:
        l_space.append(np.linspace(min_max[0], min_max[1], steps))

    return np.meshgrid(*l_space)


def _theory_predict(
    cycle: Cycle, conditions: Sequence, predict_proba: bool = False
) -> list:
    """
    Gets theory predictions over conditions space and saves results of each cycle to a list.
    Args:
        cycle: AER Cycle object that has been run
        conditions: Condition space. Should be an array of grouped conditions.
        predict_proba: Use estimator.predict_proba method instead of estimator.predict.

    Returns: list

    """
    l_predictions = []
    for i, theory in enumerate(cycle.data.theories):
        if not predict_proba:
            l_predictions.append(theory.predict(conditions))
        else:
            l_predictions.append(theory.predict_proba(conditions))

    return l_predictions


def _check_replace_default_kw(default: dict, user: dict) -> dict:
    """
    Combines the key/value pairs of two dictionaries, a default and user dictionary. Unique pairs
    are selected and user pairs take precedent over default pairs if matching keywords. Also works
    with nested dictionaries.

    Returns: dict
    """
    # Copy dict 1 to return dict
    d_return = default.copy()
    # Loop by keys in dict 2
    for key in user.keys():
        # If not in dict 1 add to the return dict
        if key not in default.keys():
            d_return.update({key: user[key]})
        else:
            # If value is a dict, recurse to check nested dict
            if isinstance(user[key], dict):
                d_return.update(
                    {key: _check_replace_default_kw(default[key], user[key])}
                )
            # If not a dict update the default value with the value from dict 2
            else:
                d_return.update({key: user[key]})

    return d_return


def plot_results_panel_2d(
    cycle: Cycle,
    iv_name: Optional[str] = None,
    dv_name: Optional[str] = None,
    steps: int = 50,
    wrap: int = 4,
    query: Optional[Union[List, slice]] = None,
    subplot_kw: dict = {},
    scatter_previous_kw: dict = {},
    scatter_current_kw: dict = {},
    plot_theory_kw: dict = {},
) -> plt.figure:
    """
    Generates a multi-panel figure with 2D plots showing results of one AER cycle.

    Observed data is plotted as a scatter plot with the current cycle colored differently than
    observed data from previous cycles. The current cycle's theory is plotted as a line over the
    range of the observed data.

    Args:
        cycle: AER Cycle object that has been run
        iv_name: Independent variable name. Name should match the name instantiated in the cycle
                    object. Default will select the first.
        dv_name: Single dependent variable name. Name should match the names instantiated in the
                    cycle object. Default will select the first DV.
        steps: Number of steps to define the condition space to plot the theory.
        wrap: Number of panels to appear in a row. Example: 9 panels with wrap=3 results in a
                3x3 grid.
        query: Query which cycles to plot with either a List of indexes or a slice. The slice must
                be constructed with the `slice()` function or `np.s_[]` index expression.
        subplot_kw: Dictionary of keywords to pass to matplotlib 'subplot' function
        scatter_previous_kw: Dictionary of keywords to pass to matplotlib 'scatter' function that
                    plots the data points from previous cycles.
        scatter_current_kw: Dictionary of keywords to pass to matplotlib 'scatter' function that
                    plots the data points from the current cycle.
        plot_theory_kw: Dictionary of keywords to pass to matplotlib 'plot' function that plots the
                    theory line.

    Returns: matplotlib figure

    """

    # ---Figure and plot params---
    # Set defaults, check and add user supplied keywords
    # Default keywords
    subplot_kw_defaults = {
        "gridspec_kw": {"bottom": 0.16},
        "sharex": True,
        "sharey": True,
    }
    scatter_previous_defaults = {
        "color": "black",
        "s": 2,
        "alpha": 0.6,
        "label": "Previous Data",
    }
    scatter_current_defaults = {
        "color": "tab:orange",
        "s": 2,
        "alpha": 0.6,
        "label": "New Data",
    }
    line_kw_defaults = {"label": "Theory"}
    # Combine default and user supplied keywords
    d_kw = {}
    for d1, d2, key in zip(
        [
            subplot_kw_defaults,
            scatter_previous_defaults,
            scatter_current_defaults,
            line_kw_defaults,
        ],
        [subplot_kw, scatter_previous_kw, scatter_current_kw, plot_theory_kw],
        ["subplot_kw", "scatter_previous_kw", "scatter_current_kw", "plot_theory_kw"],
    ):
        assert isinstance(d1, dict)
        assert isinstance(d2, dict)
        d_kw[key] = _check_replace_default_kw(d1, d2)

    # ---Extract IVs and DV metadata and indexes---
    ivs, dvs = _get_variable_index(cycle)
    if iv_name:
        iv = [s for s in ivs if s[1] == iv_name][0]
    else:
        iv = [ivs[0]][0]
    if dv_name:
        dv = [s for s in dvs if s[1] == dv_name][0]
    else:
        dv = [dvs[0]][0]
    iv_label = f"{iv[1]} {iv[2]}"
    dv_label = f"{dv[1]} {dv[2]}"

    # Create a dataframe of observed data from cycle
    df_observed = _observed_to_df(cycle)

    # Generate IV space
    condition_space = _generate_condition_space(cycle, steps=steps)

    # Get theory predictions over space
    l_predictions = _theory_predict(cycle, condition_space)

    # Cycle Indexing
    cycle_idx = list(range(len(cycle.data.theories)))
    if query:
        if isinstance(query, list):
            cycle_idx = [cycle_idx[s] for s in query]
        elif isinstance(query, slice):
            cycle_idx = cycle_idx[query]

    # Subplot configurations
    n_cycles_to_plot = len(cycle_idx)
    if n_cycles_to_plot < wrap:
        shape = (1, n_cycles_to_plot)
    else:
        shape = (int(np.ceil(n_cycles_to_plot / wrap)), wrap)
    fig, axs = plt.subplots(*shape, **d_kw["subplot_kw"])
    # Place axis object in an array if plotting single panel
    if shape == (1, 1):
        axs = np.array([axs])

    # Loop by panel
    for i, ax in enumerate(axs.flat):
        if i + 1 <= n_cycles_to_plot:
            # Get index of cycle to plot
            i_cycle = cycle_idx[i]

            # ---Plot observed data---
            # Independent variable values
            x_vals = df_observed.loc[:, iv[0]]
            # Dependent values masked by current cycle vs previous data
            dv_previous = np.ma.masked_where(
                df_observed["cycle"] >= i_cycle, df_observed[dv[0]]
            )
            dv_current = np.ma.masked_where(
                df_observed["cycle"] != i_cycle, df_observed[dv[0]]
            )
            # Plotting scatter
            ax.scatter(x_vals, dv_previous, **d_kw["scatter_previous_kw"])
            ax.scatter(x_vals, dv_current, **d_kw["scatter_current_kw"])

            # ---Plot Theory---
            conditions = condition_space[:, iv[0]]
            ax.plot(conditions, l_predictions[i_cycle], **d_kw["plot_theory_kw"])

            # Label Panels
            ax.text(
                0.05, 1, f"Cycle {i_cycle}", ha="left", va="top", transform=ax.transAxes
            )

        else:
            ax.axis("off")

    # Super Labels
    fig.supxlabel(iv_label, y=0.07)
    fig.supylabel(dv_label)

    # Legend
    fig.legend(
        ["Previous Data", "New Data", "Theory"],
        ncols=3,
        bbox_to_anchor=(0.5, 0),
        loc="lower center",
    )

    return fig


def plot_results_panel_3d(
    cycle: Cycle,
    iv_names: Optional[List[str]] = None,
    dv_name: Optional[str] = None,
    steps: int = 50,
    wrap: int = 4,
    view: Optional[Tuple[float, float]] = None,
    subplot_kw: dict = {},
    scatter_previous_kw: dict = {},
    scatter_current_kw: dict = {},
    surface_kw: dict = {},
) -> plt.figure:
    """
    Generates a multi-panel figure with 3D plots showing results of one AER cycle.

    Observed data is plotted as a scatter plot with the current cycle colored differently than
    observed data from previous cycles. The current cycle's theory is plotted as a line over the
    range of the observed data.

    Args:

        cycle: AER Cycle object that has been run
        iv_names: List of up to 2 independent variable names. Names should match the names
                    instantiated in the cycle object. Default will select up to the first two.
        dv_name: Single DV name. Name should match the names instantiated in the cycle object.
                    Default will select the first DV
        steps: Number of steps to define the condition space to plot the theory.
        wrap: Number of panels to appear in a row. Example: 9 panels with wrap=3 results in a
                3x3 grid.
        view: Tuple of elevation angle and azimuth to change the viewing angle of the plot.
        subplot_kw: Dictionary of keywords to pass to matplotlib 'subplot' function
        scatter_previous_kw: Dictionary of keywords to pass to matplotlib 'scatter' function that
                    plots the data points from previous cycles.
        scatter_current_kw: Dictionary of keywords to pass to matplotlib 'scatter' function that
                    plots the data points from the current cycle.
        surface_kw: Dictionary of keywords to pass to matplotlib 'plot_surface' function that plots
                    the theory plane.

    Returns: matplotlib figure

    """
    n_cycles = len(cycle.data.theories)

    # ---Figure and plot params---
    # Set defaults, check and add user supplied keywords
    # Default keywords
    subplot_kw_defaults = {
        "subplot_kw": {"projection": "3d"},
    }
    scatter_previous_defaults = {"color": "black", "s": 2, "label": "Previous Data"}
    scatter_current_defaults = {"color": "tab:orange", "s": 2, "label": "New Data"}
    surface_kw_defaults = {"alpha": 0.5, "label": "Theory"}
    # Combine default and user supplied keywords
    d_kw = {}
    for d1, d2, key in zip(
        [
            subplot_kw_defaults,
            scatter_previous_defaults,
            scatter_current_defaults,
            surface_kw_defaults,
        ],
        [subplot_kw, scatter_previous_kw, scatter_current_kw, surface_kw],
        ["subplot_kw", "scatter_previous_kw", "scatter_current_kw", "surface_kw"],
    ):
        assert isinstance(d1, dict)
        assert isinstance(d2, dict)
        d_kw[key] = _check_replace_default_kw(d1, d2)

    # ---Extract IVs and DV metadata and indexes---
    ivs, dvs = _get_variable_index(cycle)
    if iv_names:
        iv = [s for s in ivs if s[1] == iv_names]
    else:
        iv = ivs[:2]
    if dv_name:
        dv = [s for s in dvs if s[1] == dv_name][0]
    else:
        dv = [dvs[0]][0]
    iv_labels = [f"{s[1]} {s[2]}" for s in iv]
    dv_label = f"{dv[1]} {dv[2]}"

    # Create a dataframe of observed data from cycle
    df_observed = _observed_to_df(cycle)

    # Generate IV Mesh Grid
    x1, x2 = _generate_mesh_grid(cycle, steps=steps)

    # Get theory predictions over space
    l_predictions = _theory_predict(cycle, np.column_stack((x1.ravel(), x2.ravel())))

    # Subplot configurations
    if n_cycles < wrap:
        shape = (1, n_cycles)
    else:
        shape = (int(np.ceil(n_cycles / wrap)), wrap)

    fig, axs = plt.subplots(*shape, **d_kw["subplot_kw"])

    # Loop by panel
    for i, ax in enumerate(axs.flat):
        if i + 1 <= n_cycles:

            # ---Plot observed data---
            # Independent variable values
            l_x = [df_observed.loc[:, s[0]] for s in iv]
            # Dependent values masked by current cycle vs previous data
            dv_previous = np.ma.masked_where(
                df_observed["cycle"] >= i, df_observed[dv[0]]
            )
            dv_current = np.ma.masked_where(
                df_observed["cycle"] != i, df_observed[dv[0]]
            )
            # Plotting scatter
            ax.scatter(*l_x, dv_previous, **d_kw["scatter_previous_kw"])
            ax.scatter(*l_x, dv_current, **d_kw["scatter_current_kw"])

            # ---Plot Theory---
            ax.plot_surface(
                x1, x2, l_predictions[i].reshape(x1.shape), **d_kw["surface_kw"]
            )
            # ---Labels---
            # Title
            ax.set_title(f"Cycle {i}")

            # Axis
            ax.set_xlabel(iv_labels[0])
            ax.set_ylabel(iv_labels[1])
            ax.set_zlabel(dv_label)

            # Viewing angle
            if view:
                ax.view_init(*view)

        else:
            ax.axis("off")

    # Legend
    handles, labels = axs.flatten()[0].get_legend_handles_labels()
    legend_elements = [
        handles[0],
        handles[1],
        Patch(facecolor=handles[2].get_facecolors()[0]),
    ]
    fig.legend(
        handles=legend_elements,
        labels=labels,
        ncols=3,
        bbox_to_anchor=(0.5, 0),
        loc="lower center",
    )

    return fig


def cycle_default_score(cycle: Cycle, x_vals: np.ndarray, y_true: np.ndarray):
    """
    Calculates score for each cycle using the estimator's default scorer.
    Args:
        cycle: AER Cycle object that has been run
        x_vals: Test dataset independent values
        y_true: Test dataset dependent values

    Returns:
        List of scores by cycle
    """
    l_scores = [s.score(x_vals, y_true) for s in cycle.data.theories]
    return l_scores


def cycle_specified_score(
    scorer: Callable, cycle: Cycle, x_vals: np.ndarray, y_true: np.ndarray, **kwargs
):
    """
    Calculates score for each cycle using specified sklearn scoring function.
    Args:
        scorer: sklearn scoring function
        cycle: AER Cycle object that has been run
        x_vals: Test dataset independent values
        y_true: Test dataset dependent values
        **kwargs: Keyword arguments to send to scoring function

    Returns:

    """
    # Get predictions
    if "y_pred" in inspect.signature(scorer).parameters.keys():
        l_y_pred = _theory_predict(cycle, x_vals, predict_proba=False)
    elif "y_score" in inspect.signature(scorer).parameters.keys():
        l_y_pred = _theory_predict(cycle, x_vals, predict_proba=True)

    # Score each cycle
    l_scores = []
    for y_pred in l_y_pred:
        l_scores.append(scorer(y_true, y_pred, **kwargs))

    return l_scores


def plot_cycle_score(
    cycle: Cycle,
    X: np.ndarray,
    y_true: np.ndarray,
    scorer: Optional[Callable] = None,
    x_label: str = "Cycle",
    y_label: Optional[str] = None,
    figsize: Tuple[float, float] = rcParams["figure.figsize"],
    ylim: Optional[Tuple[float, float]] = None,
    xlim: Optional[Tuple[float, float]] = None,
    scorer_kw: dict = {},
    plot_kw: dict = {},
) -> plt.Figure:
    """
    Plots scoring metrics of cycle's theories given test data.
    Args:
        cycle: AER Cycle object that has been run
        X: Test dataset independent values
        y_true: Test dataset dependent values
        scorer: sklearn scoring function (optional)
        x_label: Label for x-axis
        y_label: Label for y-axis
        figsize: Optional figure size tuple in inches
        ylim: Optional limits for the y-axis as a tuple (lower, upper)
        xlim: Optional limits for the x-axis as a tuple (lower, upper)
        scorer_kw: Dictionary of keywords for scoring function if scorer is supplied.
        plot_kw: Dictionary of keywords to pass to matplotlib 'plot' function.

    Returns:
        matplotlib.figure.Figure
    """

    # Use estimator's default scoring method if specific scorer is not supplied
    if scorer is None:
        l_scores = cycle_default_score(cycle, X, y_true)
    else:
        l_scores = cycle_specified_score(scorer, cycle, X, y_true, **scorer_kw)

    # Plotting
    fig, ax = plt.subplots(figsize=figsize)
    ax.plot(np.arange(len(cycle.data.theories)), l_scores, **plot_kw)

    # Adjusting axis limits
    if ylim:
        ax.set_ylim(*ylim)
    if xlim:
        ax.set_xlim(*xlim)

    # Labeling
    ax.set_xlabel(x_label)
    if y_label is None:
        if scorer is not None:
            y_label = scorer.__name__
        else:
            y_label = "Score"
    ax.set_ylabel(y_label)
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))

    return fig



import copy
from collections.abc import Mapping
from dataclasses import dataclass, replace
from typing import Callable, Dict, Iterable, List, Optional

import numpy as np
from sklearn.base import BaseEstimator

from autora.experimentalist.pipeline import Pipeline
from autora.utils.dictionary import LazyDict
from autora.variable import VariableCollection


@dataclass(frozen=True)
class SimpleCycleData:
    """An object passed between and updated by processing steps in the SimpleCycle."""

    # Static
    metadata: VariableCollection

    # Aggregates each cycle from the:
    # ... Experimentalist
    conditions: List[np.ndarray]
    # ... Experiment Runner
    observations: List[np.ndarray]
    # ... Theorist
    theories: List[BaseEstimator]


def _get_cycle_properties(data: SimpleCycleData):
    """
    Examples:
        Even with an empty data object, we can initialize the dictionary,
        >>> cycle_properties = _get_cycle_properties(SimpleCycleData(metadata=VariableCollection(),
        ...     conditions=[], observations=[], theories=[]))

        ... but it will raise an exception if a value isn't yet available when we try to use it
        >>> cycle_properties["%theories[-1]%"] # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        IndexError: list index out of range

        Nevertheless, we can iterate through its keys no problem:
        >>> [key for key in cycle_properties.keys()] # doctest: +NORMALIZE_WHITESPACE
        ['%observations.ivs[-1]%', '%observations.dvs[-1]%', '%observations.ivs%',
        '%observations.dvs%', '%theories[-1]%', '%theories%']

    """

    n_ivs = len(data.metadata.independent_variables)
    n_dvs = len(data.metadata.dependent_variables)
    cycle_property_dict = LazyDict(
        {
            "%observations.ivs[-1]%": lambda: data.observations[-1][:, 0:n_ivs],
            "%observations.dvs[-1]%": lambda: data.observations[-1][:, n_ivs:],
            "%observations.ivs%": lambda: np.row_stack(
                [np.empty([0, n_ivs + n_dvs])] + data.observations
            )[:, 0:n_ivs],
            "%observations.dvs%": lambda: np.row_stack(data.observations)[:, n_ivs:],
            "%theories[-1]%": lambda: data.theories[-1],
            "%theories%": lambda: data.theories,
        }
    )
    return cycle_property_dict


class SimpleCycle:
    """
    Runs an experimentalist, theorist and experiment runner in a loop.

    Once initialized, the `cycle` can be started using the `cycle.run` method
        or by calling `next(cycle)`.

    The `.data` attribute is updated with the results.

    Attributes:
        data (dataclass): an object which is updated during the cycle and has the following
            properties:

            - `metadata`
            - `conditions`: a list of np.ndarrays representing all of the IVs proposed by the
                experimentalist
            - `observations`: a list of np.ndarrays representing all of the IVs and DVs returned by
                the experiment runner
            - `theories`: a list of all the fitted theories (scikit-learn compatible estimators)

        params (dict): a nested dictionary with parameters for the cycle parts.

                `{
                    "experimentalist": {<experimentalist params...>},
                    "theorist": {<theorist params...>},
                    "experiment_runner": {<experiment_runner params...>}
                }`


    Examples:

        ### Basic Usage

        Aim: Use the SimpleCycle to recover a simple ground truth theory from noisy data.

        >>> def ground_truth(x):
        ...     return x + 1

        The space of allowed x values is the integers between 0 and 10 inclusive,
        and we record the allowed output values as well.
        >>> from autora.variable import VariableCollection, Variable
        >>> metadata_0 = VariableCollection(
        ...    independent_variables=[Variable(name="x1", allowed_values=range(11))],
        ...    dependent_variables=[Variable(name="y", value_range=(-20, 20))],
        ...    )

        The experimentalist is used to propose experiments.
        Since the space of values is so restricted, we can just sample them all each time.
        >>> from autora.experimentalist.pipeline import make_pipeline
        >>> example_experimentalist = make_pipeline(
        ...     [metadata_0.independent_variables[0].allowed_values])

        When we run a synthetic experiment, we get a reproducible noisy result:
        >>> import numpy as np
        >>> def get_example_synthetic_experiment_runner():
        ...     rng = np.random.default_rng(seed=180)
        ...     def runner(x):
        ...         return ground_truth(x) + rng.normal(0, 0.1, x.shape)
        ...     return runner
        >>> example_synthetic_experiment_runner = get_example_synthetic_experiment_runner()
        >>> example_synthetic_experiment_runner(np.ndarray([1]))
        array([2.04339546])

        The theorist "tries" to work out the best theory.
        We use a trivial scikit-learn regressor.
        >>> from sklearn.linear_model import LinearRegression
        >>> example_theorist = LinearRegression()

        We initialize the SimpleCycle with the metadata describing the domain of the theory,
        the theorist, experimentalist and experiment runner,
        as well as a monitor which will let us know which cycle we're currently on.
        >>> cycle = SimpleCycle(
        ...     metadata=metadata_0,
        ...     theorist=example_theorist,
        ...     experimentalist=example_experimentalist,
        ...     experiment_runner=example_synthetic_experiment_runner,
        ...     monitor=lambda data: print(f"Generated {len(data.theories)} theories"),
        ... )
        >>> cycle # doctest: +ELLIPSIS
        <simple.SimpleCycle object at 0x...>

        We can run the cycle by calling the run method:
        >>> cycle.run(num_cycles=3)  # doctest: +ELLIPSIS
        Generated 1 theories
        Generated 2 theories
        Generated 3 theories
        <simple.SimpleCycle object at 0x...>

        We can now interrogate the results. The first set of conditions which went into the
        experiment runner were:
        >>> cycle.data.conditions[0]
        array([ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10])

        The observations include the conditions and the results:
        >>> cycle.data.observations[0]
        array([[ 0.        ,  0.92675345],
               [ 1.        ,  1.89519928],
               [ 2.        ,  3.08746571],
               [ 3.        ,  3.93023943],
               [ 4.        ,  4.95429102],
               [ 5.        ,  6.04763988],
               [ 6.        ,  7.20770574],
               [ 7.        ,  7.85681519],
               [ 8.        ,  9.05735823],
               [ 9.        , 10.18713406],
               [10.        , 10.88517906]])

        In the third cycle (index = 2) the first and last values are different again:
        >>> cycle.data.observations[2][[0,-1]]
        array([[ 0.        ,  1.08559827],
               [10.        , 11.08179553]])

        The best fit theory after the first cycle is:
        >>> cycle.data.theories[0]
        LinearRegression()

        >>> def report_linear_fit(m: LinearRegression,  precision=4):
        ...     s = f"y = {np.round(m.coef_[0].item(), precision)} x " \\
        ...     f"+ {np.round(m.intercept_.item(), 4)}"
        ...     return s
        >>> report_linear_fit(cycle.data.theories[0])
        'y = 1.0089 x + 0.9589'

        The best fit theory after all the cycles, including all the data, is:
        >>> report_linear_fit(cycle.data.theories[-1])
        'y = 0.9989 x + 1.0292'

        This is close to the ground truth theory of x -> (x + 1)

        We can also run the cycle with more control over the execution flow:
        >>> next(cycle) # doctest: +ELLIPSIS
        Generated 4 theories
        <simple.SimpleCycle object at 0x...>

        >>> next(cycle) # doctest: +ELLIPSIS
        Generated 5 theories
        <simple.SimpleCycle object at 0x...>

        >>> next(cycle) # doctest: +ELLIPSIS
        Generated 6 theories
        <simple.SimpleCycle object at 0x...>

        We can continue to run the cycle as long as we like,
        with a simple arbitrary stopping condition like the number of theories generated:
        >>> from itertools import takewhile
        >>> _ = list(takewhile(lambda c: len(c.data.theories) < 9, cycle))
        Generated 7 theories
        Generated 8 theories
        Generated 9 theories

        ... or the precision (here we keep iterating while the difference between the gradients
        of the second-last and last cycle is larger than 1x10^-3).
        >>> _ = list(
        ...         takewhile(
        ...             lambda c: np.abs(c.data.theories[-1].coef_.item() -
        ...                            c.data.theories[-2].coef_.item()) > 1e-3,
        ...             cycle
        ...         )
        ...     )
        Generated 10 theories
        Generated 11 theories

        ... or continue to run as long as we like:
        >>> _ = cycle.run(num_cycles=100) # doctest: +ELLIPSIS
        Generated 12 theories
        ...
        Generated 111 theories

        ### Passing Static Parameters

        It's easy to pass parameters to the cycle components, if there are any needed.
        Here we have an experimentalist which takes a parameter:
        >>> uniform_random_rng = np.random.default_rng(180)
        >>> def uniform_random_sampler(n):
        ...     return uniform_random_rng.uniform(low=0, high=11, size=n)
        >>> example_experimentalist_with_parameters = make_pipeline([uniform_random_sampler])

        The cycle can handle that using the `params` keyword:
        >>> cycle_with_parameters = SimpleCycle(
        ...     metadata=metadata_0,
        ...     theorist=example_theorist,
        ...     experimentalist=example_experimentalist_with_parameters,
        ...     experiment_runner=example_synthetic_experiment_runner,
        ...     params={"experimentalist": {"uniform_random_sampler": {"n": 7}}}
        ... )
        >>> _ = cycle_with_parameters.run()
        >>> cycle_with_parameters.data.conditions[-1].flatten()
        array([6.33661987, 7.34916618, 6.08596494, 2.28566582, 1.9553974 ,
               5.80023149, 3.27007909])

        For the next cycle, if we wish, we can change the parameter value:
        >>> cycle_with_parameters.params["experimentalist"]["uniform_random_sampler"]\\
        ...     ["n"] = 2
        >>> _ = cycle_with_parameters.run()
        >>> cycle_with_parameters.data.conditions[-1].flatten()
        array([10.5838232 ,  9.45666031])

        ### Accessing "Cycle Properties"

         Some experimentalists, experiment runners and theorists require access to the values
            created during the cycle execution, e.g. experimentalists which require access
            to the current best theory or the observed data. These data update each cycle, and
            so cannot easily be set using simple `params`.

        For this case, it is possible to use "cycle properties" in the `params` dictionary. These
            are the following strings, which will be replaced during execution by their respective
            current values:

        - `"%observations.ivs[-1]%"`: the last observed independent variables
        - `"%observations.dvs[-1]%"`: the last observed dependent variables
        - `"%observations.ivs%"`: all the observed independent variables,
            concatenated into a single array
        - `"%observations.dvs%"`: all the observed dependent variables,
            concatenated into a single array
        - `"%theories[-1]%"`: the last fitted theorist
        - `"%theories%"`: all the fitted theorists

        In the following example, we use the `"observations.ivs"` cycle property for an
            experimentalist which excludes those conditions which have
            already been seen.

        >>> metadata_1 = VariableCollection(
        ...    independent_variables=[Variable(name="x1", allowed_values=range(10))],
        ...    dependent_variables=[Variable(name="y")],
        ...    )
        >>> random_sampler_rng = np.random.default_rng(seed=180)
        >>> def custom_random_sampler(conditions, n):
        ...     sampled_conditions = random_sampler_rng.choice(conditions, size=n, replace=False)
        ...     return sampled_conditions
        >>> def exclude_conditions(conditions, excluded_conditions):
        ...     remaining_conditions = list(set(conditions) - set(excluded_conditions.flatten()))
        ...     return remaining_conditions
        >>> unobserved_data_experimentalist = make_pipeline([
        ...     metadata_1.independent_variables[0].allowed_values,
        ...     exclude_conditions,
        ...     custom_random_sampler
        ...     ]
        ... )
        >>> cycle_with_cycle_properties = SimpleCycle(
        ...     metadata=metadata_1,
        ...     theorist=example_theorist,
        ...     experimentalist=unobserved_data_experimentalist,
        ...     experiment_runner=example_synthetic_experiment_runner,
        ...     params={
        ...         "experimentalist": {
        ...             "exclude_conditions": {"excluded_conditions": "%observations.ivs%"},
        ...             "custom_random_sampler": {"n": 1}
        ...         }
        ...     }
        ... )

        Now we can run the cycler to generate conditions and run experiments. The first time round,
        we have the full set of 10 possible conditions to select from, and we select "2" at random:
        >>> _ = cycle_with_cycle_properties.run()
        >>> cycle_with_cycle_properties.data.conditions[-1]
        array([2])

        We can continue to run the cycler, each time we add more to the list of "excluded" options:
        >>> _ = cycle_with_cycle_properties.run(num_cycles=5)
        >>> cycle_with_cycle_properties.data.conditions
        [array([2]), array([6]), array([5]), array([7]), array([3]), array([4])]

        By using the monitor callback, we can investigate what's going on with the cycle properties:
        >>> cycle_with_cycle_properties.monitor = lambda data: print(
        ...     _get_cycle_properties(data)["%observations.ivs%"].flatten()
        ... )

        The monitor evaluates at the end of each cycle
        and shows that we've added a new observed IV each step
        >>> _ = cycle_with_cycle_properties.run()
        [2. 6. 5. 7. 3. 4. 9.]
        >>> _ = cycle_with_cycle_properties.run()
        [2. 6. 5. 7. 3. 4. 9. 0.]

        We deactivate the monitor by making it "None" again.
        >>> cycle_with_cycle_properties.monitor = None

        We can continue until we've sampled all of the options:
        >>> _ = cycle_with_cycle_properties.run(num_cycles=2)
        >>> cycle_with_cycle_properties.data.conditions # doctest: +NORMALIZE_WHITESPACE
        [array([2]), array([6]), array([5]), array([7]), array([3]), \
        array([4]), array([9]), array([0]), array([8]), array([1])]

        If we try to evaluate it again, the experimentalist fails, as there aren't any more
        conditions which are available:
        >>> cycle_with_cycle_properties.run()  # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        ValueError: a cannot be empty unless no samples are taken

    """

    def __init__(
        self,
        metadata: VariableCollection,
        theorist,
        experimentalist,
        experiment_runner,
        monitor: Optional[Callable[[SimpleCycleData], None]] = None,
        params: Optional[Dict] = None,
    ):
        """
        Args:
            metadata: a description of the dependent and independent variables
            theorist: a scikit-learn-compatible estimator
            experimentalist: an autora.experimentalist.Pipeline
            experiment_runner: a function to map independent variables onto observed dependent
                variables
            monitor: a function which gets read-only access to the `data` attribute at the end of
                each cycle.
            params: a nested dictionary with parameters to be passed to the parts of the cycle.
                E.g. if the experimentalist had a step named "pool" which took an argument "n",
                which you wanted to set to the value 30, then params would be set to this:
                `{"experimentalist": {"pool": {"n": 30}}}`
        """

        self.theorist = theorist
        self.experimentalist = experimentalist
        self.experiment_runner = experiment_runner
        self.monitor = monitor
        if params is None:
            params = dict()
        self.params = params

        self.data = SimpleCycleData(
            metadata=metadata,
            conditions=[],
            observations=[],
            theories=[],
        )

    def run(self, num_cycles: int = 1):
        for i in range(num_cycles):
            next(self)
        return self

    def __next__(self):
        assert (
            "experiment_runner" not in self.params
        ), "experiment_runner cannot yet accept cycle properties"
        assert (
            "theorist" not in self.params
        ), "theorist cannot yet accept cycle properties"

        data = self.data
        params_with_cycle_properties = _resolve_cycle_properties(
            self.params, _get_cycle_properties(self.data)
        )

        data = self._experimentalist_callback(
            self.experimentalist,
            data,
            params_with_cycle_properties.get("experimentalist", dict()),
        )
        data = self._experiment_runner_callback(self.experiment_runner, data)
        data = self._theorist_callback(self.theorist, data)
        self._monitor_callback(data)
        self.data = data

        return self

    def __iter__(self):
        return self

    @staticmethod
    def _experimentalist_callback(
        experimentalist: Pipeline, data_in: SimpleCycleData, params: dict
    ):
        new_conditions = experimentalist(**params)
        if isinstance(new_conditions, Iterable):
            # If the pipeline gives us an iterable, we need to make it into a concrete array.
            # We can't move this logic to the Pipeline, because the pipeline doesn't know whether
            # it's within another pipeline and whether it should convert the iterable to a
            # concrete array.
            new_conditions_values = list(new_conditions)
            new_conditions_array = np.array(new_conditions_values)
        else:
            raise NotImplementedError(f"Object {new_conditions} can't be handled yet.")

        assert isinstance(
            new_conditions_array, np.ndarray
        )  # Check the object is bounded
        data_out = replace(
            data_in,
            conditions=data_in.conditions + [new_conditions_array],
        )
        return data_out

    @staticmethod
    def _experiment_runner_callback(
        experiment_runner: Callable, data_in: SimpleCycleData
    ):
        x = data_in.conditions[-1]
        y = experiment_runner(x)
        new_observations = np.column_stack([x, y])
        data_out = replace(
            data_in, observations=data_in.observations + [new_observations]
        )
        return data_out

    @staticmethod
    def _theorist_callback(theorist, data_in: SimpleCycleData):
        all_observations = np.row_stack(data_in.observations)
        n_xs = len(
            data_in.metadata.independent_variables
        )  # The number of independent variables
        x, y = all_observations[:, :n_xs], all_observations[:, n_xs:]
        if y.shape[1] == 1:
            y = y.ravel()
        new_theorist = copy.deepcopy(theorist)
        new_theorist.fit(x, y)
        data_out = replace(
            data_in,
            theories=data_in.theories + [new_theorist],
        )
        return data_out

    def _monitor_callback(self, data: SimpleCycleData):
        if self.monitor is not None:
            self.monitor(data)


def _resolve_cycle_properties(params: Dict, cycle_properties: Mapping):
    """
    Resolve "cycle properties" inside a nested dictionary.

    In this context, a "cycle property" is a string which is meant to be replaced by a
    different value before the dictionary is used.

    Args:
        params: a (nested) dictionary of keys and values, where some values might be
            "cycle property names"
        cycle_properties: a dictionary of "cycle property names" and their "real values"

    Returns: a (nested) dictionary where "cycle property names" are replaced by the "real values"

    Examples:

        >>> params_0 = {"key": "%foo%"}
        >>> cycle_properties_0 = {"%foo%": 180}
        >>> _resolve_cycle_properties(params_0, cycle_properties_0)
        {'key': 180}

        >>> params_1 = {"key": "%bar%", "nested_dict": {"inner_key": "%foobar%"}}
        >>> cycle_properties_1 = {"%bar%": 1, "%foobar%": 2}
        >>> _resolve_cycle_properties(params_1, cycle_properties_1)
        {'key': 1, 'nested_dict': {'inner_key': 2}}

    """
    params_ = copy.copy(params)
    for key, value in params_.items():
        if isinstance(value, dict):
            params_[key] = _resolve_cycle_properties(value, cycle_properties)
        elif (
            isinstance(value, str) and value in cycle_properties
        ):  # value is a key in the cycle_properties dictionary
            params_[key] = cycle_properties[value]
        else:
            pass  # no change needed

    return params_



import random

import numpy as np


def random_pool(*args, n=1, duplicates=True):
    """
    Creates combinations from lists of discrete values using random selection.
    Args:
        *args: m lists of discrete values. One value will be sampled from each list.
        n: Number of samples to sample
        duplicates: Boolean if duplicate value are allowed.

    """
    l_samples = []
    # Create list of pools of values sample from
    pools = [tuple(pool) for pool in args]

    # Check to ensure infinite search won't occur if duplicates not allowed
    if not duplicates:
        l_pool_len = [len(set(s)) for s in pools]
        n_combinations = np.product(l_pool_len)
        try:
            assert n <= n_combinations
        except AssertionError:
            raise AssertionError(
                f"Number to sample n({n}) is larger than the number "
                f"of unique combinations({n_combinations})."
            )

    # Random sample from the pools until n is met
    while len(l_samples) < n:
        l_samples.append(tuple(map(random.choice, pools)))
        if not duplicates:
            l_samples = [*set(l_samples)]

    return iter(l_samples)
